/* -*- Mode: IDL; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is written by Juan Lang.
 *
 * The Initial Developer of the Original Code is 
 * Juan Lang.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Max Christian
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#ifdef XP_WIN

// WINVER needs to be >= 0x0500 in order for HDEVNOTIFY and other types to be
// defined.  I dynamically load (using LoadLibrary) all platform-specific
// functions, so this shouldn't make Mozilla require Win2k, just let me compile
#if(WINVER < 0x0500)
#undef WINVER
#define WINVER 0x0500
#endif /* WINVER */

#include <stdarg.h>
#include <windef.h>
#include <winbase.h>
#include <winuser.h>
#include <dbt.h>
#include <iphlpapi.h>
#endif
#include "nspr.h"
#include "nsCRT.h"
#include "nsAutoLock.h"
#include "nsIEventQueueService.h"
#include "nsIObserverService.h"
#include "nsIServiceManager.h"
#include "nsNetInterfaceDetector.h"

#ifdef XP_WIN
// jil TODO:  these should be prefs?
#define CARD_INSERTION_POLL_MS 100
#define CARD_INSERTION_MIN_WAIT_MS 30 * 1000
#define CARD_INSERTION_MAX_WAIT_MS  3 * 60 * 1000

typedef DWORD (WINAPI *GetAdaptersInfoFunc)(PIP_ADAPTER_INFO, PULONG);
typedef DWORD (WINAPI *GetIfEntryFunc)(PMIB_IFROW);
typedef HDEVNOTIFY (WINAPI *RegisterDeviceNotificationFunc)(HANDLE, LPVOID,
 DWORD);
typedef BOOL (WINAPI *UnregisterDeviceNotificationFunc)(HDEVNOTIFY);

static HMODULE gIPHelper = nsnull;
static GetAdaptersInfoFunc gGetAdaptersInfo = nsnull;
static GetIfEntryFunc gGetIfEntry = nsnull;

static void initIPHelperLibrary(void);
static void freeIPHelperLibrary(void);
#endif /* XP_WIN */

NS_IMPL_THREADSAFE_ISUPPORTS3(nsNetInterfaceDetector, nsINetInterfaceDetector, nsIRunnable, nsIObserver)

nsNetInterfaceDetector *nsNetInterfaceDetector::mInstance = nsnull;

nsNetInterfaceDetector::nsNetInterfaceDetector()
    : mThread(0)
    , mLock(nsnull)
    , mCondVar(nsnull)
{
    NS_ASSERTION(!mInstance, "nsNetInterfaceDetector already initialized");
    mInstance = this;
#ifdef XP_WIN
    mWindow = 0;
    mHaveAtLeastOneInterface = PR_FALSE;
    mWaitState = Initial;
    initIPHelperLibrary();
#endif /* XP_WIN */
}

nsNetInterfaceDetector::~nsNetInterfaceDetector()
{
    nsresult rv = ShutdownInternal();
    NS_ASSERTION(NS_SUCCEEDED(rv), "NetInterfaceDetector shutdown failed");
    NS_ASSERTION(!mThread, "NetInterfaceDetector thread shutdown failed");
    NS_ASSERTION(!mLock, "NetInterfaceDetector lock shutdown failed");
    NS_ASSERTION(!mCondVar,
     "NetInterfaceDetector condition variable not destroyed");
    NS_ASSERTION(mInstance == this,
     "destroying nsNetInterfaceDetector other than singleton");
    mInstance = nsnull;
#ifdef XP_WIN
    freeIPHelperLibrary();
#endif /* XP_WIN */
}

nsresult
nsNetInterfaceDetector::Init()
{
    nsresult rv;
    nsCOMPtr<nsIObserverService> observerService = 
     do_GetService("@mozilla.org/observer-service;1", &rv);

    if (NS_FAILED(rv)) return rv;
    rv = observerService->AddObserver(this, NS_XPCOM_SHUTDOWN_OBSERVER_ID, PR_FALSE);
    if (NS_FAILED(rv)) return rv;

#ifdef XP_WIN
    mLock = PR_NewLock();
    if (!mLock)
        return NS_ERROR_OUT_OF_MEMORY;
    mCondVar = PR_NewCondVar(mLock);
    if (mCondVar) {
        nsAutoLock lock(mLock);

        rv = NS_NewThread(getter_AddRefs(mThread), this, 0, PR_JOINABLE_THREAD);
        NS_ASSERTION(NS_SUCCEEDED(rv), "failed to create device notification thread");
        // sync with notification thread to allow it to create its window
        while (!mWindow)
            PR_WaitCondVar(mCondVar, PR_INTERVAL_NO_TIMEOUT);
    }
    else {
        PR_DestroyLock(mLock);
        rv = NS_ERROR_OUT_OF_MEMORY;
    }
#endif

    return rv;
}

#ifdef XP_WIN
NS_IMETHODIMP
nsNetInterfaceDetector::HaveNetworkInterface(PRBool *_retval)
{
    if (!_retval) return NS_ERROR_NULL_POINTER;
    *_retval = InternalHaveNetworkInterface(nsnull);
    return NS_OK;
}
#else
NS_IMETHODIMP
nsNetInterfaceDetector::HaveNetworkInterface(PRBool *_retval)
{
    if (!_retval) return NS_ERROR_NULL_POINTER;
    // note to UNIX porters:  I considered using SIOCGIFCONF and SIOCGIFADDR
    // to check for network addresses, but this would only tell me if I have
    // an IPv4 interface.  I don't know a good XPlat way to check for IPv6
    // addresses, and I don't actually use IPv6 so code I wrote would be
    // suspect anyway.  Suggestions welcome.
    *_retval = PR_TRUE;
    return NS_OK;
}
#endif /* XP_WIN */

#if defined (XP_WIN)
NS_IMETHODIMP
nsNetInterfaceDetector::Run()
{
    WNDCLASS wc;
    char szWindowClass[] = "Mozilla:DevNotificationClass";

    memset(&wc, 0, sizeof(wc));
    wc.lpfnWndProc   = WndProc;
    wc.lpszClassName = szWindowClass;
    RegisterClassA(&wc);

    mWindow = CreateWindowA(szWindowClass, "Mozilla:DevNotificationWindow",
     0, 0, 0, 10, 10, NULL, NULL, NULL, NULL);

    PR_Lock(mLock);
    PR_NotifyCondVar(mCondVar);
    PR_Unlock(mLock);

    HDEVNOTIFY hDevNotify = 0;
    HMODULE user32 = LoadLibraryA("user32.dll");
    if (user32) {
        RegisterDeviceNotificationFunc registerDeviceNotificationA =
         (RegisterDeviceNotificationFunc)GetProcAddress(user32,
         "RegisterDeviceNotificationA");
        if (registerDeviceNotificationA) {
            DEV_BROADCAST_DEVICEINTERFACE_A devInterface;
            // from ndisguid.h in NTDDK, aka GUID_NDIS_LAN_CLASS
            GUID netGUID = {
             0xad498944, 0x762f, 0x11d0, 0x8d, 0xcb, 0x00, 0xc0, 0x4f, 0xc3,
             0x35, 0x8c };

            memset(&devInterface, 0, sizeof(devInterface));
            devInterface.dbcc_size = sizeof(devInterface);
            devInterface.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
            devInterface.dbcc_classguid = netGUID;
            hDevNotify = registerDeviceNotificationA(mWindow, &devInterface,
             DEVICE_NOTIFY_WINDOW_HANDLE);
        }
    }

    // I run the message loop whether or not RegisterDeviceNotification
    // succeeds, because I *may* get device notifications without a successful
    // registration on Win98 (not sure)
    MSG msg;
    while (GetMessage(&msg, mWindow, 0, 0)) {
        // no TranslateMessage() because we're not expecting input
        DispatchMessage(&msg);
    }

    if (user32) {
        if (hDevNotify) {
            OSVERSIONINFO info;

            // don't call UnregisterDeviceNotification on Win98, as it can
            // flake out the OS--bastards
            info.dwOSVersionInfoSize = sizeof(info);
            GetVersionEx(&info);
            if (info.dwPlatformId == VER_PLATFORM_WIN32_NT) {
                UnregisterDeviceNotificationFunc unregisterDeviceNotification =
                 (UnregisterDeviceNotificationFunc)GetProcAddress(user32,
                 "UnregisterDeviceNotification");

                if (unregisterDeviceNotification)
                    unregisterDeviceNotification(hDevNotify);
            }
        }
        FreeLibrary(user32);
    }

    UnregisterClassA(szWindowClass, NULL);

    return NS_OK;
}
#else
NS_IMETHODIMP
nsNetInterfaceDetector::Run()
{
    // note to porters:  it should be possible to detect network card
    // appearance/disappearance events on certain platforms.
    // E.g. on MacOSX, see Living in a Dynamic TCP/IP Environment,
    // http://developer.apple.com/technotes/tn/tn1145.html
    // On Linux, /sbin/hotplug sends notifications, see
    // http://linux-hotplug.sourceforge.net/?selected=net
    return NS_OK;
}
#endif /* XP_WIN */

NS_IMETHODIMP
nsNetInterfaceDetector::Observe(nsISupports *subject,
                                const char *topic,
                                const PRUnichar *data)
{
    nsresult rv;

    if (!nsCRT::strcmp(NS_XPCOM_SHUTDOWN_OBSERVER_ID, topic))
        rv = ShutdownInternal();
    else
        rv = NS_OK;
    return rv;
}

/* Sends the given event to the UI thread.  Assumes aEventID never goes out
 * of scope (static strings are ideal).
 */
nsresult
nsNetInterfaceDetector::SendEventToUI(const char *aEventID)
{
    nsresult rv;

    if (!aEventID) return NS_ERROR_NULL_POINTER;

    nsCOMPtr<nsIEventQueueService> eqs =
      do_GetService(NS_EVENTQUEUESERVICE_CONTRACTID);
    if (!eqs) {
        NS_ERROR("failed to get EventQueue service");
        return NS_ERROR_FAILURE;
    }

    nsCOMPtr<nsIEventQueue> eq;
    rv = eqs->GetThreadEventQueue(NS_UI_THREAD, getter_AddRefs(eq));
    if (NS_FAILED(rv))
        return rv;

    MyEvent *event = new MyEvent(aEventID);
    if (!event)
        return NS_ERROR_OUT_OF_MEMORY;
    // AddRef this because it is being placed in the PLEvent; it'll be Released
    // when DestroyInterfaceEvent is called
    NS_ADDREF(mInstance);
    PL_InitEvent(event, mInstance, HandleInterfaceEvent, DestroyInterfaceEvent);

    if (eq->PostEvent(event) == PR_FAILURE) {
        NS_ERROR("failed to post event to UI EventQueue");
        NS_RELEASE(mInstance);
        delete event;
        rv = NS_ERROR_FAILURE;
    }
    return rv;
}

void * PR_CALLBACK
nsNetInterfaceDetector::HandleInterfaceEvent(PLEvent *aEvent)
{
    MyEvent *event = NS_STATIC_CAST(MyEvent *, aEvent);

    nsCOMPtr<nsIObserverService> observerService =
      do_GetService("@mozilla.org/observer-service;1");
    // don't care if notification fails
    if (observerService)
        observerService->NotifyObservers(
         NS_STATIC_CAST(nsINetInterfaceDetector *, mInstance),
         event->mEventID, NULL);
    return nsnull;
}

void PR_CALLBACK
nsNetInterfaceDetector::DestroyInterfaceEvent(PLEvent *aEvent)
{
    nsNetInterfaceDetector *nid =
     (nsNetInterfaceDetector *)PL_GetEventOwner(aEvent);
    NS_IF_RELEASE(nid);
    delete aEvent;
}

nsresult
nsNetInterfaceDetector::ShutdownInternal()
{
    nsresult rv;

#if defined (XP_WIN)
    SendMessage(mWindow, WM_DESTROY, 0, 0);
#endif /* XP_WIN */
    if (mThread)
        rv = mThread->Join();

    if (mLock)
        PR_Lock(mLock);

    // remove xpcom shutdown observer
    nsCOMPtr<nsIObserverService> observerService =
     do_GetService("@mozilla.org/observer-service;1", &rv);
    if (NS_SUCCEEDED(rv))
        rv = observerService->RemoveObserver(this,
         NS_XPCOM_SHUTDOWN_OBSERVER_ID);

    // Have to break the cycle here, otherwise nsNetInterfaceDetector holds
    // onto the thread and the thread holds onto the nsNetInterfaceDetector
    // via its mRunnable
    mThread = nsnull;

    if (mCondVar) {
        PR_DestroyCondVar(mCondVar);
        mCondVar = nsnull;
    }

    if (mLock) {
        PR_Unlock(mLock);
        PR_DestroyLock(mLock);
        mLock = nsnull;
    }

    return rv;
}

#ifdef XP_WIN
static void initIPHelperLibrary(void)
{
    if (!gIPHelper)
        gIPHelper = LoadLibraryA("iphlpapi.dll");
    if (gIPHelper) {
        gGetAdaptersInfo = (GetAdaptersInfoFunc)GetProcAddress(gIPHelper,
         "GetAdaptersInfo");
        gGetIfEntry = (GetIfEntryFunc)GetProcAddress(gIPHelper,
         "GetIfEntry");
    }
}

static void freeIPHelperLibrary(void)
{
    if (gIPHelper) {
        gGetAdaptersInfo = nsnull;
        gGetIfEntry = nsnull;
        FreeLibrary(gIPHelper);
        gIPHelper = nsnull;
    }
}

VOID NS_STDCALL
nsNetInterfaceDetector::TimerCallback(HWND hWnd, UINT msg, UINT_PTR idEvent,
 DWORD dwTime)
{
    NS_ASSERTION(mInstance, "missing nsNetInterfaceDetector singleton");
    if (mInstance) {
        PRBool haveInterface, waitingOnDHCP;

        haveInterface = mInstance->InternalHaveNetworkInterface(&waitingOnDHCP);
        if (haveInterface) {
            KillTimer(hWnd, idEvent);
            SendEventToUI(NS_NETWORK_INTERFACE_APPEARED_OBSERVER_ID);
        }
        else {
            if (dwTime > mInstance->mStopPollingTime) {
                if (mInstance->mWaitState == Initial && waitingOnDHCP) {
                    mInstance->mWaitState = DHCP;
                    mInstance->mStopPollingTime = GetTickCount() +
                     CARD_INSERTION_MAX_WAIT_MS - CARD_INSERTION_MIN_WAIT_MS;
                }
                else
                    KillTimer(hWnd, idEvent);
            }
        }
    }
}

LRESULT NS_STDCALL
nsNetInterfaceDetector::WndProc(HWND hWnd, UINT uMsg, WPARAM wp, LPARAM lp)
{
    LRESULT retVal = 0;

    switch (uMsg) 
    {
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    case WM_DEVICECHANGE:
        switch (wp)
        {
        case DBT_DEVICEARRIVAL:
            if (mInstance && !mInstance->mHaveAtLeastOneInterface) {
                // new card arrived, maybe I can be online now, but delay
                // checking because link state may not be right yet
                mInstance->mStopPollingTime = GetTickCount() +
                 CARD_INSERTION_MIN_WAIT_MS;
                SetTimer(hWnd, 0, CARD_INSERTION_POLL_MS, TimerCallback);
            }
            retVal = TRUE;
            break;
        case DBT_DEVICEREMOVECOMPLETE:
            // no delay needed here, b/c a disappeared card isn't going to have
            // a valid link state ever
            if (mInstance && !mInstance->InternalHaveNetworkInterface(nsnull))
                SendEventToUI(NS_NETWORK_INTERFACE_DISAPPEARED_OBSERVER_ID);
            retVal = TRUE;
            break;
        default:
            retVal = TRUE;
        }
        break;
    default:
        retVal = DefWindowProc(hWnd, uMsg, wp, lp);
    }
    return retVal;
}

/* Returns whether there's at least one working network interface.  As a side
 * effect, remembers this fact in mHaveAtLeastOneInterface.  Optionally can
 * tell you whether not having one is due to an adapter that's configured for
 * DHCP but doesn't have an IP address yet.  (If you don't care, pass nsnull
 * for aWaitingOnDHCP.)  Experience shows this isn't always reliable--it takes
 * some time before the IpHelper API sets the DhcpEnabled value.  You should
 * probably poll for a little while before trusting aWaitingOnDHCP.
 */
PRBool
nsNetInterfaceDetector::InternalHaveNetworkInterface(PRBool *aWaitingOnDHCP)
{
    if (aWaitingOnDHCP)
        *aWaitingOnDHCP = PR_FALSE;
    if (gGetAdaptersInfo && gGetIfEntry) {
        ULONG adaptersLen = 0;
        PIP_ADAPTER_INFO adapters = nsnull;

        mHaveAtLeastOneInterface = PR_FALSE;
        gGetAdaptersInfo(0, &adaptersLen);
        adapters = (PIP_ADAPTER_INFO)PR_Malloc(adaptersLen);
        if (gGetAdaptersInfo(adapters, &adaptersLen) == ERROR_SUCCESS) {
            PIP_ADAPTER_INFO ptr = nsnull;

            for (ptr = adapters; ptr && !mHaveAtLeastOneInterface;
             ptr = ptr->Next) {
                MIB_IFROW ifRow;

                ifRow.dwIndex = ptr->Index;
                if (gGetIfEntry(&ifRow) == NO_ERROR && ifRow.dwOperStatus >=
                 MIB_IF_OPER_STATUS_CONNECTED) {
                    PIP_ADDR_STRING ipAddr;
                    PRBool hasValidAddress = PR_FALSE;

                    for (ipAddr = &ptr->IpAddressList; ipAddr &&
                     !hasValidAddress; ipAddr = ipAddr->Next) {
                        // this appears to be IPv4-specific, but the special
                        // string value I'm comparing to should only be set
                        // with DHCP (v4) enabled on an adapter, before an
                        // address is configured.
                        if (strcmp(ipAddr->IpAddress.String, "0.0.0.0") != 0)
                            hasValidAddress = PR_TRUE;
                    }
                    if (hasValidAddress)
                        mHaveAtLeastOneInterface = PR_TRUE;
                    else {
                        if (ptr->DhcpEnabled && aWaitingOnDHCP)
                            *aWaitingOnDHCP = PR_TRUE;
                    }
                }
            }
        }
        PR_Free(adapters);
    }
    else {
        // library not loaded, so assume we have an adapter
        mHaveAtLeastOneInterface = PR_TRUE;
    }
    return mHaveAtLeastOneInterface;
}
#endif /* XP_WIN */
